"""This script tests using EPICS CA and Python threads together

  Based on code from  Friedrich Schotte, NIH
  modified by Matt Newville 19-Apr-2010

  modified MN, 22-April-2011 (1 year later!)
  to support new context-switching modes

"""

import time
import epics
import sys
from threading import Thread
from epics.ca import CAThread, withInitialContext
from  pvnames import updating_pvlist
write = sys.stdout.write
flush = sys.stdout.flush


pvs_b = []
names_b = []
for pvname in updating_pvlist:
    names_b.append(pvname)

names_a = names_b[1:]
pvs_a   = pvs_b[1:]

epics.ca.create_context()

def wait_for_changes(pvnames, runtime, runname):
    """basic test procedure called by other tests
    """
    def onChanges(pvname=None, value=None, char_value=None, **kw):
        write('   %s= %s (%s)\n' % (pvname, char_value, runname))
        flush()
    t0 = time.time()
    pvs = []
    for pvn in pvnames:
        p = epics.PV(pvn)
        p.get()
        p.add_callback(onChanges)
        pvs.append(p)

    while time.time()-t0 < runtime:
        try:
            time.sleep(0.01)
        except:
            sys.exit()

    for p in pvs:
        p.clear_callbacks()

def run_initcontext(pvnames, runtime, run_name):
    write(' -> force inital ca context: thread=%s will run for %.3f sec\n' % (run_name, runtime))
    epics.ca.use_initial_context()
    wait_for_changes(pvnames, runtime, run_name)
    write( 'Done with Thread  %s\n' % ( run_name))

@withInitialContext
def run_decorator(pvnames, runtime, run_name):
    write(' -> use withInitialContext decorator: thread=%s will run for %.3f sec\n' % (run_name, runtime))
    wait_for_changes(pvnames, runtime, run_name)
    write( 'Done with Thread  %s\n' % ( run_name))

def run_CAThread(pvnames, runtime, run_name):
    write(' -> used with CAThread: thread=%s will run for %.3f sec\n' % (run_name, runtime))
    wait_for_changes(pvnames, runtime, run_name)
    write( 'Done with Thread  %s\n' % ( run_name))

def run_threads(threadlist):
    for th in threadlist:
        th.start()
    time.sleep(0.01)
    for th in threadlist:
        th.join()
    time.sleep(0.01)

def test_initcontext():
    write( 'Test use plain threading.Thread, force use of initial CA Context \n')
    th1 = Thread(target=run_initcontext, args=(names_a, 2, 'A'))
    th2 = Thread(target=run_initcontext, args=(names_b, 3, 'B'))
    run_threads((th1, th2))

def test_decorator():
    write( 'Test use plain threading.Thread, withInitialContext decorator\n')
    th1 = Thread(target=run_decorator, args=(names_a, 3, 'A'))
    th2 = Thread(target=run_decorator, args=(names_b, 5, 'B'))
    run_threads((th1, th2))

def test_cathread():
    write( 'Test use CAThread\n')
    th1 = CAThread(target=run_CAThread, args=(names_a, 3, 'A'))
    th2 = CAThread(target=run_CAThread, args=(names_b, 5, 'B'))
    run_threads((th1, th2))



import epics
import threading
import pvnames


def test_basic_thread():
    result = []
    def thread():
        epics.ca.use_initial_context()
        pv = epics.get_pv(pvnames.double_pv)
        result.append(pv.get())

    epics.ca.use_initial_context()
    t = threading.Thread(target=thread)
    t.start()
    t.join()

    assert len(result) and result[0] is not None


def test_basic_cathread():
    result = []
    def thread():
        pv = epics.get_pv(pvnames.double_pv)
        result.append(pv.get())

    epics.ca.use_initial_context()
    t = epics.ca.CAThread(target=thread)
    t.start()
    t.join()

    assert len(result) and result[0] is not None


def test_attach_context():
    result = []
    def thread():
        epics.ca.create_context()
        pv = epics.get_pv(pvnames.double_pv2)
        assert pv.wait_for_connection()
        result.append(pv.get())
        epics.ca.detach_context()

        epics.ca.attach_context(ctx)
        pv = epics.get_pv(pvnames.double_pv)
        assert pv.wait_for_connection()
        result.append(pv.get())

    epics.ca.use_initial_context()
    ctx = epics.ca.current_context()
    t = threading.Thread(target=thread)
    t.start()
    t.join()

    assert len(result) == 2 and result[0] is not None
    print(result)


def test_pv_from_main():
    result = []
    def thread():
        result.append(pv.get())

    epics.ca.use_initial_context()
    pv = epics.get_pv(pvnames.double_pv2)

    t = epics.ca.CAThread(target=thread)
    t.start()
    t.join()

    assert len(result) and result[0] is not None



from __future__ import print_function

import time
import gc
import os
import epics

# a test for possible memory leaks on put()
import pvnames

pvlist = pvnames.char_arrays #  + pvnames.long_arrays + pvnames.double_arrays

def show_memory():
    gc.collect()
    if os.name == 'nt':
        return 'Windows memory usage?? pid=%i' % os.getpid()
    f = open("/proc/%i/statm" % os.getpid())
    mem = f.readline().split()
    f.close()
    return 'Memory: VmSize = %i kB  /  VmRss = %i kB' %( int(mem[0])*4 , int(mem[1])*4)

N_new = 0
def get_callback(pv=None, **kws):
    global N_new
    N_new = N_new + 1
    # print( 'New value: ', pv.pvname, pv.char_value)

def monitor_events(t = 600.0):
    print('Processing PV requests:')
    t0 = time.time()
    endtime = t0 + t
    nx = 0
    global N_new
    nnotify = int(t / 30)
    while time.time() < endtime:
        epics.ca.pend_event(0.05)
        nx  = nx + 1
        if nx >=nnotify:
            print("changes (%i) / %.3f /  %s" % (N_new,  time.time()-t0, show_memory()))
            N_new = 0
            nx = 0

pvs = [epics.PV(i, callback=get_callback) for i in pvlist]
epics.ca.pend_io()

for i in range(500):
    for p in pvs:
        p.put('test: run %i' % (i))
    epics.ca.pend_event(0.02)
    if i%20 == 0:
        print("==run #  ", i,  show_memory())
    time.sleep(0.02)

epics.ca.pend_io(1.0)

print('really done.')



from .. import Device

class AD_OverlayPlugin(Device):
    """
    AreaDetector Overlay Plugin
    """
    attrs = ('Name', 'Name_RBV',
             'Use', 'Use_RBV',
             'PositionX', 'PositionX_RBV',
             'PositionY', 'PositionY_RBV',
             'PositionXLink', 'PositionYLink', 
             'SizeXLink', 'SizeYLink', 
             'SizeX', 'SizeX_RBV',
             'SizeY', 'SizeY_RBV',
             'Shape', 'Shape_RBV',
             'DrawMode', 'DrawMode_RBV',
             'Red',    'Red_RBV',
             'Green', 'Green_RBV',
             'Blue', 'Blue_RBV')

    _nonpvs = ('_prefix', '_pvs', '_delim')

    def __init__(self, prefix):
        Device.__init__(self, prefix, delim='', mutable=False,
                        attrs=self.attrs)

    def ensure_value(self, attr, value, wait=False):
        """ensures that an attribute with an associated _RBV value is
        set to the specifed value
        """
        rbv_attr = "%s_RBV" % attr
        if rbv_attr not in self._pvs:
            return self._pvs[attr].put(value, wait=wait)

        if  self._pvs[rbv_attr].get(as_string=True) != value:
            self._pvs[attr].put(value, wait=wait)




import time
from . import ca, dbr, pv, alarm, autosave, device, motor, multiproc
from .version import __version__

__doc__ = f"""
   Epics Channel Access Python module

   version: {__version__}
   Principal Authors:
      Matthew Newville <newville@cars.uchicago.edu> CARS, University of Chicago
      Ken Lauer, David Chabot, Angus Gratton

== License:
   Except where explicitly noted, this file and all files in this distribution
   are licensed under the Epics Open License See LICENSE in the top-level
   directory of this distribution.

== Overview:
   Python Interface to the Epics Channel Access protocol of the Epics control system.
"""

PV = pv.PV
Alarm = alarm.Alarm
Motor = motor.Motor
Device = device.Device
poll = ca.poll

get_pv = pv.get_pv

CAProcess = multiproc.CAProcess
CAPool = multiproc.CAPool

# some constants
NO_ALARM = 0
MINOR_ALARM = 1
MAJOR_ALARM = 2
INVALID_ALARM = 3

_PVmonitors_ = {}

def caput(pvname, value, wait=False, timeout=60.0, connection_timeout=5.0):
    """caput(pvname, value, wait=False, timeout=60, connection_timeout=5.0)

    put a value to an epics Process Variable (PV).

    Arguments
    ---------
     pvname (str):   name of PV
     value (any):    value to put to PV (see notes)
     wait (bool):    whether to wait for processing to complete [False]
     timeout (float):  maximum time to wait for the processing to complete [60]
     connection_timeout (float): maximum time to wait for connection [5]

    Returns
    --------
      1  on succesful completion
      -1 or other negative value on failure of the low-level CA put.
      None  on a failure to connect to the PV.

    Notes
    ------
      1. Epics PVs are typically limited to an appropriate Epics data type,
         int, float, str, and enums or homogeneously typed lists or arrays.
         Numpy arrays or Python strings are generally coeced as appropriate,
         but care may be needed when mapping Python objects to Epics PV values.
      2. If not already connected, the PV will first attempt to connect to
         the networked variable. As the initial connection can take some time
         (typically 30 msec or so), if a successful connection is made, a rich
         PV object with will be stored internally for later use.  Use
         connection_timeout to control how long to wait before declaring that
         the PV cannot be connected (mispelled name, inactive IOC, improper
         network configuration)
      3. Since some PVs can take a long time to process (perhaps physically
         moving a motor or telling a detector to collect and not return until
         done), it is impossible to tell what a "reasonable" timeout for a put
         should be.
      4. All time in seconds.

    Examples
    ---------
       to put a value to a PV and return as soon as possible:
       >>> caput('xx.VAL', 3.0)

      to wait for processing to finish, use 'wait=True':
      >>> caput('xx.VAL', 3.0, wait=True)

    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=connection_timeout, connect=True)
    out = None
    if thispv.connected:
        timeout -= time.time() - start_time
        out = thispv.put(value, wait=wait, timeout=timeout)
    return out


def caget(pvname, as_string=False, count=None, as_numpy=True,
          use_monitor=True, timeout=5.0, connection_timeout=5.0):
    """caget(pvname, as_string=False, count=None, as_numpy=True,
             use_monitor=True, timeout=5.0, connection_timeout=5.0)

    get the current value to an epics Process Variable (PV).

    Arguments
    ---------
     pvname (str):   name of PV
     as_string (bool): whether to get the string representation [False]
     count (int or None): maximum number of elements of array values [None]
     use_monitor (bool): whether to use the value cached by the monitor [True]
     timeout (float):  maximum time to wait for the processing to complete [60]
     connection_timeout (float): maximum time to wait for connection [5]

    Returns
    --------
      the PV value on success, or `None`on a failure.

    Notes
    ------
      1. If not already connected, the PV will first attempt to connect to
         the networked variable. As the initial connection can take some time
         (typically 30 msec or so), if a successful connection is made, a rich
         PV object with will be stored internally for later use.
      2. as_string=True will return a string: float values will formatted
         according to the PVs precision, enum values will return the approriate
         enum string, etc.
      3. count can be used to limit the number of elements fetched for array PVs.
      4. `use_monitor=False` will return the most recently cached value from the
         internal monitor placed on the PV. This will be the latest value unless
         the value is changing very rapidly. `use_monitor=False` will ignore the
         cached value and ask for an explicit value.  Of course, if a PV is
         changing rapidly enough for that to make a difference, it may also
         change between getting the value and downstream code using it.
      5. All time values are in seconds.


    Examples
    ---------
       get of a PV's value
       >>> x = caget('xx.VAL')

       to get the character string representation (formatted double,
       enum string, etc):
       >>> x = caget('xx.VAL', as_string=True)

       to get a truncated amount of data from an array, you can specify
       the count with
       >>> x = caget('MyArray.VAL', count=1000)
    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=connection_timeout, connect=True)
    val = None
    if thispv.connected:
        if as_string:
            thispv.get_ctrlvars()
        timeout -= time.time() - start_time
        val = thispv.get(count=count, timeout=timeout,
                         use_monitor=use_monitor,
                         as_string=as_string,
                         as_numpy=as_numpy)
        poll()
    return val

def cainfo(pvname, print_out=True, timeout=5.0, connection_timeout=5.0):
    """cainfo(pvname,print_out=True,timeout=5.0, connection_timeout=5.0)

    return printable information about PV

    Arguments
    ---------
     pvname (str):   name of PV
     print_out (bool): whether to print the info to standard out [True]
     timeout (float):  maximum time to wait for the processing to complete [60]
     connection_timeout (float): maximum time to wait for connection [5]

    Returns
    --------
       if `print_out` is True, returns the printable text
       otherwise return `None`.

    Notes
    ------
      1. If not already connected, the PV will first attempt to connect to
         the networked variable. As the initial connection can take some time
         (typically 30 msec or so), if a successful connection is made, a rich
         PV object with will be stored internally for later use.
      2. All time in seconds.


    Examples
    ---------
       to print a status report for the PV:
       >>>cainfo('xx.VAL')

       to get the multiline text, use
       >>>txt = cainfo('xx.VAL', print_out=False)

    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=connection_timeout, connect=True)
    out = None
    if thispv.connected:
        conn_time = time.time() - start_time
        thispv.get(timeout=timeout-conn_time)
        get_time = time.time() - start_time
        thispv.get_ctrlvars(timeout=timeout-get_time)
        if print_out:
            ca.write(thispv.info)
        else:
            out = thispv.info
    return out

def camonitor_clear(pvname):
    """clear monitors on a PV"""
    if pvname in _PVmonitors_:
        _PVmonitors_[pvname].remove_callback(index=-999)
        _PVmonitors_.pop(pvname)

def camonitor(pvname, writer=None, callback=None, connection_timeout=5.0):
    """ camonitor(pvname, writer=None, callback=None, connection_timeout=5)

    sets a monitor on a PV.

    Arguments
    ---------
     pvname (str):   name of PV
     writer (callable or None): function used to send monitor messages [None]
     callback (callback or None): custom callback function [None]
     connection_timeout (float): maximum time to wait for connection [5]


    Notes
    ------
     1. To write the result to a file, provide the writer option a
        write method to an open file or some other callable that
        accepts a string.
     2. To completely control where the output goes, provide a callback
         method and you can do whatever you'd like with them.
     3. A custom callback can be provieded.  This will be sent keyword
        arguments for pvname, value, char_value, and more.
        Important: use **kws!!

    Examples
    --------
      to write a message with the latest value for that PV each
      time the value changes and when ca.poll() is called.
      >>>camonitor('xx.VAL')
    """

    if writer is None:
        writer = ca.write
    if callback is None:
        def callback(pvname=None, value=None, char_value=None, **kwds):
            "generic monitor callback"
            if char_value is None:
                char_value = repr(value)
            writer(f"{pvname:.32s} {pv.fmt_time()} {char_value}")

    thispv = get_pv(pvname, timeout=connection_timeout, connect=True)
    if thispv.connected:
        thispv.get()
        thispv.add_callback(callback, index=-999, with_ctrlvars=True)
        _PVmonitors_[pvname] = thispv

def caget_many(pvlist, as_string=False, as_numpy=True, count=None,
               timeout=1.0, connection_timeout=5.0, conn_timeout=None):
    """caget_many(pvlist, as_string=False, as_numpy=True, count=None,
                 timeout=1.0, connection_timeout=5.0, conn_timeout=None)
    get values for a list of PVs, working as fast as possible

    Arguments
    ---------
     pvlist (list):        list of pv names to fetch
     as_string (bool):     whether to get values as strings [False]
     as_numpy (bool):      whether to get values as numpy arrys [True]
     count  (int or None): max number of elements to get [None]
     timeout (float):      timeout on *each* get()  [1.0]
     connection_timeout (float): timeout for *all* pvs to connect [5.0]
     conn_timeout (float):  back-compat alias or connection_timeout

    Returns
    --------
      list of values, with `None` signifying 'not connected' or 'timed out'.

    Notes
    ------
       this does not cache PV objects.

    """
    chids, connected, out = [], [], []
    for name in pvlist:
        chids.append(ca.create_channel(name, auto_cb=False, connect=False))

    all_connected = False
    if conn_timeout is not None:
        connection_timeout = conn_timeout
    expire_time = time.time() + connection_timeout
    while (not all_connected and (time.time() < expire_time)):
        connected = [dbr.CS_CONN==ca.state(chid) for chid in chids]
        all_connected = all(connected)
        poll()

    for (chid, conn) in zip(chids, connected):
        if conn:
            ca.get(chid, count=count, as_string=as_string, as_numpy=as_numpy,
                   wait=False)

    poll()
    for (chid, conn) in zip(chids, connected):
        val = None
        if conn:
            val = ca.get_complete(chid, count=count, as_string=as_string,
                                  as_numpy=as_numpy, timeout=timeout)
        out.append(val)
    return out

def caput_many(pvlist, values, wait=False, connection_timeout=5.0,
               put_timeout=60):
    """caput_many(pvlist, values, wait=False, connection_timeout=5.0,
               put_timeout=60)


    put values to a list of PVs, as quickly as possible

    Arguments
    ----------
     pvlist (iterable):   list of PV names
     values (iterable):   list of values for corresponding PVs
     wait  (bool or string):   whether to wait for puts (see notes) [False]
     put_timeout (float):  maximum time to wait for the put to complete [60]
     connection_timeout (float): maximum time to wait for connection [5]


    Returns
    --------
     a list of ints or `None`, with values of 1 if the put was successful,
     or a negative number if the put failed (say, the timeout was exceeded),
     or `None` if the connection failed.

    Notes
    ------
      1. This does not maintain the PV objects.

      2. If wait is 'each', *each* put operation will block until it is
         complete or until the put_timeout duration expires.
         If wait is 'all', this method will block until *all* put
         operations are complete, or until the put_timeout expires.

         This 'wait' only applies to the put timeout, not the
         connection timeout.

    """
    if len(pvlist) != len(values):
        raise ValueError("List of PV names must be equal to list of values.")
    kwargs = {'auto_monitor': False, 'timeout': connection_timeout}
    pvs = [get_pv(name, **kwargs) for name in pvlist]

    wait_all = wait=='all'
    put_kws = {'wait': (wait=='each'), 'timeout': put_timeout,
               'use_complete': wait_all}
    put_ret = []
    for pvo, val in zip(pvs, values):
        put_ret.append(pvo.put(val, **put_kws))
    out = None
    if wait_all:
        start_time = time.time()
        while not all(((pv.connected and pv.put_complete) for pv in pvs)):
            ca.poll()
            elapsed_time = time.time() - start_time
            if elapsed_time > put_timeout:
                break
        out = [1 if (pv.connected and pv.put_complete) else -1 for pv in pvs]
    else:
        out = [val if val == 1 else -1 for val in put_ret]
    return out



"""
String and data utils
"""
import sys
import os

IOENCODING =  os.environ.get('PYEPICS_ENCODING',
              os.environ.get('PYTHONIOENCODING',
                             'utf-8'))

def str2bytes(st1):
    'string to byte conversion'
    if isinstance(st1, bytes):
        return st1
    return bytes(st1, IOENCODING)

def bytes2str(st1):
    'byte to string conversion'
    if isinstance(st1, str):
        return st1
    elif isinstance(st1, bytes):
        return str(st1, IOENCODING)
    else:
        return str(st1)

def strjoin(sep, seq):
    "join string sequence with a separator"
    if isinstance(sep, bytes):
        sep = bytes2str(sep)
    if len(seq) == 0:
        seq = ''
    elif isinstance(seq[0], bytes):
        tmp =[]
        for i in seq:
            if i == b'\x00':
                break
            tmp.append(bytes2str(i))
        seq = tmp
    return sep.join(seq)


def clib_search_path(lib):
    '''Assemble path to c library.

    Parameters
    ----------
    lib : str
        Either 'ca' or 'Com'.

    Returns
    --------
    str : string

    Examples
    --------
    >>> clib_search_path('ca')
    'linux64/libca.so'

    '''

    # determine which libca / libCom dll is appropriate
    try:
        import platform
        nbits = platform.architecture()[0]
        mach = platform.machine()
    except:
        nbits = '32bit'
        mach = 'x86_64'

    nbits = nbits.replace('bit', '')
    if mach.startswith('arm'):
        nbits = 'arm'

    libfmt = 'lib%s.so'
    if os.name == 'nt':
        libsrc = 'win'
        libfmt = '%s.dll'
    elif sys.platform == 'darwin':
        libsrc = 'darwin'
        if platform.machine() == 'arm64':
            libsrc, nbits = 'darwinarm', '64'
        libfmt = 'lib%s.dylib'
    elif sys.platform.startswith('linux'):
        libsrc = 'linux'

    else:
        return None

    return os.path.join("%s%s" % (libsrc, nbits), libfmt % lib)



#!/usr/bin/env python

import epics
import time
import sys
import getopt

def show_usage():
    s = """Usage caget.py [options] <PV name> ...
    -h:    show this message


    """
    print s
    sys.exit()
    
opts, args = getopt.getopt(sys.argv[1:], "htansd:#:e:f:g:w:",
                           ["help", "terse","wide","number","data=",])

wait_time = 1.0

arr_num = None
format = None
terse  = False
use_ts = False
dbr_type = None
enum_as_num = False

for (k,v) in opts:
    if k in ("-h", "--help"):   show_usage()
    elif k == '-t':     terse = True
    elif k == '-a':    use_ts = True
    elif k == '-n':    enum_as_num = True
    elif k == '-d':    dbr_type = int(v)
    elif k == '-#':    arr_num  = int(v)
    elif k == '-w':    wait_time= float(v)
    elif k == '-e':     format = '%%.%ie' % int(v)
    elif k == '-f':      format = '%%.%if' % int(v)
    elif k == '-g':     format = '%%.%ig' % int(v)
    elif k == '-s':     format = 'STRING'

for pvname in args:
    form='native'
    if use_ts: form = 'time'
    pv = epics.PV(pvname,form=form)

    # pv connection
    t0 = time.time()
    pv.connect()
    while time.time()-t0 < wait_time:
        if pv.connected: break
        epics.poll()
    if not pv.connected:
        print "%s:: not connected" % pvname
    ox = pv.get_ctrlvars()
    epics.poll()


    value = pv.get(as_string=format=='STRING')

    if pv.type == 'enum' and  not enum_as_num:
        enum_strs = pv.enum_strs
        value = enum_strs[pv.get()]

    outstring = ''
    if not terse: outstring = pvname

    if use_ts:
        ts = epics.dbr.EPICS2UNIX_EPOCH +  pv._args['timestamp']
        outstring = "%s   %s" %(outstring,time.ctime(ts))
        print ts
    print outstring, value
    



import epics
import epics.ca
import epics.pv

import time
import subprocess
import pvnames
enabledpv = pvnames.clear_cache_enabled
beaconspv = pvnames.clear_cache_beacons


change_count = 0


def onChanges(pvname=None, value=None, **kw):
    global change_count
    change_count += 1


def run():
    for beaconpv in beaconspv:
        value = epics.caget(beaconpv)
        assert value is not None, \
            "PV {} is offline".format(beaconpv)
    time.sleep(0.2)
    epics.ca.clear_cache()


def test_use_monitor():
    # This test verifies that clear_cache clears subscriptions created by
    # use_monitor=True prior clearing channels.
    # Otherwise, EPICS will crash on an attempt to clear subscription
    # on a cleared channel.
    assert epics.caget(enabledpv, use_monitor=True) is not None, \
        "PV {} is offline".format(enabledpv)

    pv = epics.pv.get_pv(enabledpv)

    epics.ca.clear_cache()

    pv.disconnect()


def test_subscribe():
    # This test verifies that clear_cache does not
    # leave registered callbacks behind which may cause
    # a crash on shutdown on Linux randomly.
    # Therefore, the test runs itself
    # in a new process multiple times.
    # This test takes quite long time to complete.
    assert epics.caget(enabledpv) is not None, \
        "PV {} is offline".format(enabledpv)
    epics.caput(enabledpv, 1)
    for run in range(20):
        try:
            subprocess.run(['python', __file__], check=True)
        except subprocess.CalledProcessError as err:
            epics.caput(enabledpv, 0)
            if err.returncode > 0:
                error = 'the return code {}'.format(err.returncode)
            else:
                error = 'the signal {}'.format(-err.returncode)
            assert False, "Run {} failed due to {}".format(
                run,
                error)
    epics.caput(enabledpv, 0)


if (__name__ == '__main__'):
    run()



#!/usr/bin/env python
# unit-tests for ca interface

import sys
import time
import numpy
import threading
import pytest
from random import random
from contextlib import contextmanager
from epics import PV, get_pv, caput, caget, caget_many, caput_many, ca

import pvnames

def write(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()

CONN_DAT ={}
CHANGE_DAT = {}

def onConnect(pvname=None, conn=None, chid=None,  **kws):
    write('  :Connection status changed:  %s  connected=%s\n' % (pvname, conn))
    global CONN_DAT
    CONN_DAT[pvname] = conn

def onChanges(pvname=None, value=None, **kws):
    write( '/// New Value: %s  value=%s, kw=%s\n' %( pvname, str(value), repr(kws)))
    global CHANGE_DAT
    CHANGE_DAT[pvname] = value

@contextmanager
def no_simulator_updates():
    '''Context manager which pauses and resumes simulator PV updating'''
    try:
        caput(pvnames.pause_pv, 1, wait=True)
        time.sleep(0.05)
        yield
    finally:
        caput(pvnames.pause_pv, 0, wait=True)


def test_CreatePV():
    write('Simple Test: create pv\n')
    pv = get_pv(pvnames.double_pv)
    assert pv is not None

def test_CreatedWithConn():
    write('Simple Test: create pv with conn callback\n')
    pv = get_pv(pvnames.int_pv, connection_callback=onConnect)
    val = pv.get()

    global CONN_DAT
    conn = CONN_DAT.get(pvnames.int_pv, None)
    assert conn

def test_caget():
    write('Simple Test of caget() function\n')
    pvs = (pvnames.double_pv, pvnames.enum_pv, pvnames.str_pv)
    for p in pvs:
        val = caget(p)
        assert val is not None
    assert caget(pvnames.str_pv) == 'ao'

def test_caget_many():
    write('Simple Test of caget_many() function\n')
    pvs = [pvnames.double_pv, pvnames.enum_pv, pvnames.str_pv]
    vals = caget_many(pvs)
    assert len(vals) == len(pvs)
    assert isinstance(vals[0], float)
    assert isinstance(vals[1], int)
    assert isinstance(vals[2], str)

def test_caput_many_wait_all():
    write('Test of caput_many() function, waiting for all.\n')
    pvs = [pvnames.double_pv, pvnames.enum_pv, 'ceci nest pas une PV']
    #pvs = ["MTEST:Val1", "MTEST:Val2", "MTEST:SlowVal"]
    vals = [0.5, 0, 23]
    t0 = time.time()
    success = caput_many(pvs, vals, wait='all', connection_timeout=0.5, put_timeout=5.0)
    t1 = time.time()
    assert len(success) == len(pvs)
    assert success[0] == 1
    assert success[1] == 1
    assert success[2] < 0


def test_caput_many_wait_each():
    write('Simple Test of caput_many() function, waiting for each.\n')
    pvs = [pvnames.double_pv, pvnames.enum_pv, 'ceci nest pas une PV']
    #pvs = ["MTEST:Val1", "MTEST:Val2", "MTEST:SlowVal"]
    vals = [0.5, 0, 23]
    success = caput_many(pvs, vals, wait='each', connection_timeout=0.5, put_timeout=1.0)
    assert len(success) == len(pvs)
    assert success[0] == 1
    assert success[1] == 1
    assert success[2] < 0

def test_caput_many_no_wait():
    write('Simple Test of caput_many() function, without waiting.\n')
    pvs = [pvnames.double_pv, pvnames.enum_pv, 'ceci nest pas une PV']
    #pvs = ["MTEST:Val1", "MTEST:Val2", "MTEST:SlowVal"]
    vals = [0.5, 0, 23]
    success = caput_many(pvs, vals, wait=None, connection_timeout=0.5)
    assert len(success) == len(pvs)
    #If you don't wait, ca.put returns 1 as long as the PV connects
    #and the put request is valid.
    assert (success[0] == 1)
    assert (success[1] == 1)
    assert (success[2] < 0)

def test_get1():
    write('Simple Test: test value and char_value on an integer\n')
    with no_simulator_updates():
        pv = get_pv(pvnames.int_pv)
        val = pv.get()
        cval = pv.get(as_string=True)
        assert int(cval) == val

def test_get_with_metadata():
    with no_simulator_updates():
        pv = get_pv(pvnames.int_pv, form='native')

        # Request time type
        md = pv.get_with_metadata(use_monitor=False, form='time')
        assert 'timestamp' in md
        assert 'lower_ctrl_limit' not in md

        # Request control type
        md = pv.get_with_metadata(use_monitor=False, form='ctrl')
        assert 'lower_ctrl_limit' in md
        assert 'timestamp' not in md

        # Use monitor: all metadata should come through
        md = pv.get_with_metadata(use_monitor=True)
        assert 'timestamp' in md
        assert 'lower_ctrl_limit' in md

        # Get a namespace
        ns = pv.get_with_metadata(use_monitor=True, as_namespace=True)
        assert hasattr(ns, 'timestamp')
        assert hasattr(ns, 'lower_ctrl_limit')

def test_get_string_waveform():
    write('String Array: \n')
    with no_simulator_updates():
        pv = get_pv(pvnames.string_arr_pv)
        val = pv.get()
        assert (len(val) > 10)
        assert isinstance(val[0], str)
        assert len(val[0]) > 1
        assert isinstance(val[1], str)
        assert len(val[1]) > 1

def test_putcomplete():
    write('Put with wait and put_complete (using real motor!) \n')
    vals = (1.35, 1.50, 1.44, 1.445, 1.45, 1.453, 1.446, 1.447, 1.450, 1.450, 1.490, 1.5, 1.500)
    p = get_pv(pvnames.motor1)
    # this works with a real motor, fail if it doesn't connect quickly
    if not p.wait_for_connection(timeout=0.2):
        return

    see_complete = []
    for v in vals:
        t0 = time.time()
        p.put(v, use_complete=True)
        count = 0
        for i in range(100000):
            time.sleep(0.001)
            count = count + 1
            if p.put_complete:
                see_complete.append(True)
                break
            # print( 'made it to value= %.3f, elapsed time= %.4f sec (count=%i)' % (v, time.time()-t0, count))
    assert len(see_complete) > (len(vals) - 5)

def test_putwait():
    write('Put with wait (using real motor!) \n')
    pv = get_pv(pvnames.motor1)
    # this works with a real motor, fail if it doesn't connect quickly
    if not pv.wait_for_connection(timeout=0.2):
        return

    val = pv.get()

    t0 = time.time()
    if val < 5:
        pv.put(val + 1.0, wait=True)
    else:
        pv.put(val - 1.0, wait=True)
    dt = time.time()-t0
    write('    put took %s sec\n' % dt)
    assert  dt > 0.1

    # now with a callback!
    global put_callback_called
    put_callback_called = False

    def onPutdone(pvname=None, **kws):
        print( 'put done ', pvname, kws)
        global put_callback_called
        put_callback_called = True
    val = pv.get()
    if  val < 5:
        pv.put(val + 1.0, callback=onPutdone)
    else:
        pv.put(val - 1.0, callback=onPutdone)

    t0 = time.time()
    while time.time()-t0 < dt*1.50:
        time.sleep(0.02)

    write('    put should be done by now?  %s \n' % put_callback_called)
    assert put_callback_called

    # now using pv.put_complete
    val = pv.get()
    if  val < 5:
        pv.put(val + 1.0, use_complete=True)
    else:
        pv.put(val - 1.0, use_complete=True)
    t0 = time.time()
    count = 0
    while time.time()-t0 < dt*1.50:
        if pv.put_complete:
            break
        count = count + 1
        time.sleep(0.02)
    write('    put_complete=%s (should be True), and count=%i (should be>3)\n' %
          (pv.put_complete, count))
    assert pv.put_complete
    assert count > 3

def test_get_callback():
    write("Callback test:  changing PV must be updated\n")
    global NEWVALS
    mypv = get_pv(pvnames.updating_pv1)
    NEWVALS = []
    def onChanges(pvname=None, value=None, char_value=None, **kw):
        write( 'PV %s %s, %s Changed!\n' % (pvname, repr(value), char_value))
        NEWVALS.append( repr(value))

    mypv.add_callback(onChanges)
    write('Added a callback.  Now wait for changes...\n')

    t0 = time.time()
    while time.time() - t0 < 3:
        time.sleep(1.e-4)
    write('   saw %i changes.\n' % len(NEWVALS))
    assert len(NEWVALS) > 3
    mypv.clear_callbacks()

def test_put_string_waveform():
    write('String Array: put\n')
    with no_simulator_updates():
        pv = get_pv(pvnames.string_arr_pv)
        put_value = ['a', 'b', 'c']
        pv.put(put_value, wait=True)
        get_value = pv.get(use_monitor=False, as_numpy=False)
        numpy.testing.assert_array_equal(get_value, put_value)

def test_put_string_waveform_single_element():
    write('String Array: put single element\n')
    with no_simulator_updates():
        pv = get_pv(pvnames.string_arr_pv)
        put_value = ['a']
        pv.put(put_value, wait=True)
        time.sleep(0.05)
        get_value = pv.get(use_monitor=False, as_numpy=False)
        assert put_value == get_value

def test_put_string_waveform_mixed_types():
    write('String Array: put mixed types\n')
    with no_simulator_updates():
        pv = get_pv(pvnames.string_arr_pv)
        put_value = ['a', 2, 'b']
        pv.put(put_value, wait=True)
        time.sleep(0.05)
        get_value = pv.get(use_monitor=False, as_numpy=False)
        numpy.testing.assert_array_equal(get_value, ['a', '2', 'b'])

def test_put_string_waveform_empty_list():
    write('String Array: put empty list\n')
    with no_simulator_updates():
        pv = get_pv(pvnames.string_arr_pv)
        put_value = []
        pv.put(put_value, wait=True)
        time.sleep(0.05)
        get_value = pv.get(use_monitor=False, as_numpy=False)
        assert '' == ''.join(get_value)

def test_put_string_waveform_zero_length_strings():
    write('String Array: put zero length strings\n')
    with no_simulator_updates():
        pv = get_pv(pvnames.string_arr_pv)
        put_value = ['', '', '']
        pv.put(put_value, wait=True)
        time.sleep(0.05)
        get_value = pv.get(use_monitor=False, as_numpy=False)
        numpy.testing.assert_array_equal(get_value, put_value)


def test_string_waveform_lengths():
    "test string arrays with length=1 and other lengths"
    spv = get_pv(pvnames.string_arr_pv)
    readback = spv.get()
    nelm = spv.nelm
    with no_simulator_updates():
        for count in (nelm, nelm-5, nelm//3, 11, 5, 1, 3, 7, 2, 1):
            dat = [f"str_{count}_{i}" for i in range(count)]
            spv.put(dat)
            time.sleep(0.1)
            readback = spv.get()
            assert all(readback == dat)


def test_subarrays():
    write("Subarray test:  dynamic length arrays\n")
    driver = get_pv(pvnames.subarr_driver)
    subarr1 = get_pv(pvnames.subarr1)
    subarr1.connect()

    len_full = 64
    len_sub1 = 16
    full_data = numpy.arange(len_full)/1.0

    caput("%s.NELM" % pvnames.subarr1, len_sub1)
    caput("%s.INDX" % pvnames.subarr1, 0)


    driver.put(full_data) ;
    time.sleep(0.1)
    subval = subarr1.get()

    assert (len(subval) == len_sub1)
    assert numpy.all(subval == full_data[:len_sub1])
    write("Subarray test:  C\n")
    caput("%s.NELM" % pvnames.subarr2, 19)
    caput("%s.INDX" % pvnames.subarr2, 3)

    subarr2 = get_pv(pvnames.subarr2)
    subarr2.get()

    driver.put(full_data) ;   time.sleep(0.1)
    subval = subarr2.get()

    assert len(subval) == 19
    assert (numpy.all(subval == full_data[3:3+19]))

    caput("%s.NELM" % pvnames.subarr2, 5)
    caput("%s.INDX" % pvnames.subarr2, 13)

    driver.put(full_data) ;   time.sleep(0.1)
    subval = subarr2.get()

    assert len(subval) == 5
    assert (numpy.all(subval == full_data[13:5+13]))

def test_subarray_zerolen():
    subarr1 = get_pv(pvnames.zero_len_subarr1)
    subarr1.wait_for_connection()

    val = subarr1.get(use_monitor=True, as_numpy=True)
    assert isinstance(val, numpy.ndarray)
    assert len(val) == 0
    assert val.dtype == numpy.float64

    val = subarr1.get(use_monitor=False, as_numpy=True)
    assert isinstance(val, numpy.ndarray)
    assert  len(val) == 0
    assert val.dtype == numpy.float64


def test_waveform_get_with_count_arg():
    with no_simulator_updates():
        # NOTE: do not use get_pv() here, as `count` is incompatible with
        # the cache
        wf = PV(pvnames.char_arr_pv, count=32)
        val=wf.get()
        assert len(val) == 32

        val=wf.get(count=wf.nelm)
        assert len(val) == wf.nelm

def test_waveform_callback_with_count_arg():
    values = []

    # NOTE: do not use get_pv() here, as `count` is incompatible with
    # the cache
    wf = PV(pvnames.char_arr_pv, count=32)
    def onChanges(pvname=None, value=None, char_value=None, **kw):
        write( 'PV %s %s, %s Changed!\n' % (pvname, repr(value), char_value))
        values.append( value)

    wf.add_callback(onChanges)
    write('Added a callback.  Now wait for changes...\n')

    t0 = time.time()
    while time.time() - t0 < 3:
        time.sleep(1.e-4)
        if len(values)>0:
            break

    assert len(values) > 0
    assert len(values[0]) == 32

    wf.clear_callbacks()


def test_emptyish_char_waveform_no_monitor():
    '''a test of a char waveform of length 1 (NORD=1): value "\0"
    without using auto_monitor
    '''
    with no_simulator_updates():
        zerostr = PV(pvnames.char_arr_pv, auto_monitor=False)
        zerostr.wait_for_connection()

        # elem_count = 128, requested count = None, libca returns count = 1
        zerostr.put([0], wait=True)
        assert zerostr.get(as_string=True) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0])
        assert zerostr.get(as_string=True, as_numpy=False) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0])

        # elem_count = 128, requested count = None, libca returns count = 2
        zerostr.put([0, 0], wait=True)
        assert zerostr.get(as_string=True) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0, 0])
        assert zerostr.get(as_string=True, as_numpy=False) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0, 0])
        zerostr.disconnect()

def test_emptyish_char_waveform_monitor():
    '''a test of a char waveform of length 1 (NORD=1): value "\0"
    with using auto_monitor
    '''
    with no_simulator_updates():
        zerostr = PV(pvnames.char_arr_pv, auto_monitor=True)
        zerostr.wait_for_connection()

        zerostr.put([0], wait=True)
        time.sleep(0.2)

        assert zerostr.get(as_string=True) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0])
        assert zerostr.get(as_string=True, as_numpy=False) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0])

        zerostr.put([0, 0], wait=True)
        time.sleep(0.2)

        assert zerostr.get(as_string=True) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0, 0])
        assert zerostr.get(as_string=True, as_numpy=False) == ''
        numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0, 0])
        zerostr.disconnect()

def testEnumPut():
    pv = get_pv(pvnames.enum_pv)
    assert pv is not None
    pv.put('Stop')
    time.sleep(0.1)
    assert pv.get() == 0

def test_DoubleVal():
    pvn = pvnames.double_pv
    pv = get_pv(pvn)
    pv.get()
    cdict  = pv.get_ctrlvars()
    write( 'Testing CTRL Values for a Double (%s)\n'   % (pvn))
    assert 'severity' in cdict
    assert len(pv.host) > 1
    assert pv.count == 1
    assert pv.precision == pvnames.double_pv_prec
    assert pv.units == pvnames.double_pv_units
    assert pv.access.startswith('read')


def test_type_converions_2():
    write("CA type conversions arrays\n")
    pvlist = (pvnames.char_arr_pv,
              pvnames.long_arr_pv,
              pvnames.double_arr_pv)
    with no_simulator_updates():
        chids = []
        for name in pvlist:
            chid = ca.create_channel(name)
            ca.connect_channel(chid)
            chids.append((chid, name))
            ca.poll(evt=0.025, iot=5.0)
        ca.poll(evt=0.05, iot=10.0)

        values = {}
        for chid, name in chids:
            values[name] = ca.get(chid)
        for promotion in ('ctrl', 'time'):
            for chid, pvname in chids:
                write('=== %s  chid=%s as %s\n' % (ca.name(chid),
                                                   repr(chid), promotion))
                time.sleep(0.01)
                if promotion == 'ctrl':
                    ntype = ca.promote_type(chid, use_ctrl=True)
                else:
                    ntype = ca.promote_type(chid, use_time=True)

                val  = ca.get(chid, ftype=ntype)
                cval = ca.get(chid, as_string=True)
                for a, b in zip(val, values[pvname]):
                    assert a == b

def test_waveform_get_1elem():
    pv = get_pv(pvnames.double_arr_pv)
    val = pv.get(count=1, use_monitor=False)
    assert isinstance(val, numpy.ndarray)
    assert (len(val) == 1)

def test_subarray_1elem():
    with no_simulator_updates():
        # pv = get_pv(pvnames.zero_len_subarr1)
        pv = get_pv(pvnames.double_arr_pv)
        pv.wait_for_connection()

        val = pv.get(count=1, use_monitor=False)
        print('val is', val, type(val))
        assert isinstance(val, numpy.ndarray)
        assert len(val) == 1

        val = pv.get(count=1, as_numpy=False, use_monitor=False)
        print('val is', val, type(val))
        assert isinstance(val, list)
        assert len(val) == 1


@pytest.mark.parametrize('num_threads', [1, 10, 200])
@pytest.mark.parametrize('thread_class', [ca.CAThread, threading.Thread])
def test_multithreaded_get(num_threads, thread_class):
    def thread(thread_idx):
        result[thread_idx] = (pv.get(),
                              pv.get_with_metadata(form='ctrl')['value'],
                              pv.get_with_metadata(form='time')['value'],
                              )

    result = {}
    ca.use_initial_context()
    pv = get_pv(pvnames.double_pv)

    threads = [thread_class(target=thread, args=(i, ))
               for i in range(num_threads)]

    with no_simulator_updates():
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

    assert len(result) == num_threads
    print(result)
    values = set(result.values())
    assert len(values) != 0

    value, = values
    assert value is not None


@pytest.mark.parametrize('num_threads', [1, 10, 100])
def test_multithreaded_put_complete(num_threads):
    def callback(pvname, data):
        result.append(data)

    def thread(thread_idx):
        pv.put(thread_idx, callback=callback,
               callback_data=dict(data=thread_idx),
               wait=True)
        time.sleep(0.1)

    result = []
    ca.use_initial_context()
    pv = get_pv(pvnames.double_pv)

    threads = [ca.CAThread(target=thread, args=(i, ))
               for i in range(num_threads)]

    with no_simulator_updates():
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

    assert len(result) == num_threads
    print(result)
    assert set(result) == set(range(num_threads))


def test_force_connect():
    pv = get_pv(pvnames.double_arrays[0], auto_monitor=True)

    print("Connecting")
    assert pv.wait_for_connection(5.0)

    print("SUM", pv.get().sum())

    time.sleep(3)

    print("Disconnecting")
    pv.disconnect()

    called = {'called': False}
    def callback(value=None, **kwargs):
        called['called'] = True
        print("update", value.sum())

    pv.add_callback(callback)

    print("Reconnecting")
    pv.force_connect()

    assert pv.wait_for_connection(5.0)

    time.sleep(2.0)
    assert pv.get() is not None
    assert called['called']



"""
Port of Mark Rivers epicsPV class to use epics.PV

Matt Newville  7-April-2010

Original comments:

This module defines the epicsPV class, which adds additional features to
Geoff Savage's CaChannel class.

Author:         Mark Rivers
Created:        Sept. 16, 2002.
Modifications:
"""
import epics

class epicsPV(epics.PV):
    """
    This class subclasses PV to provide a compatible API to Mark Rivers
    epicsPV class
    
    - setMonitor() sets a generic callback routine for value change events.
    Subsequent getw(), getValue() or array_get() calls will return the
    value from the most recent callback, and hence do not result in any
    network activity or latency.  This can greatly improve performance.
        
    - checkMonitor() returns a flag to indicate if a callback has occured
    since the last call to checkMonitor(), getw(), getValue() or
    array_get().  It can be used to increase efficiency in polling
    applications.
        
    - getControl() reads the "control" and other information from an
    EPICS PV without having to use callbacks.
    In addition to the PV value, this will return the graphic, control and
    alarm limits, etc.

    - putWait() calls array_put_callback() and waits for the callback to
    occur before it returns.  This allows programs to use array_put_callback()
    synchronously and without user-written callbacks.
    
    Created:  Mark Rivers, Sept. 16, 2002.
    Modifications:
    """

    def __init__(self, pvname=None, wait=True):
        """
        Keywords:
        pvname:
        An optional name of an EPICS Process Variable.
        
        wait: If wait==True and pvname != None then this constructor will do
        a wait for connection for the PV.  If wait==0 and pvname != None
        then the PV will eventually connect...
        
        """
        # Invoke the base class initialization
        PV.__init__(self, pvname)
        self.monitorState = False
        if pvname is not None and wait:
            self.connect()

    def _getCallback(self, pvname=None, value=None, **kw):
        self.monitorState = True

    def setMonitor(self):
        """
        Sets a simple callback routine for value change events
        to note when a change occurs.
        """
        self.monitorState = False
        self.add_callback(callback=self._getCallback)

    def clearMonitor(self):
        """
        Cancels the effect of a previous call to setMonitor().
        """
        self.monitorState = False

    def checkMonitor(self):
        """
        Returns True to indicate if a value callback has occured,
        indicating a new value is available since the last check.
        Returns False if no such callback has occurred.
        """
        epics.poll()
        out =  self.monitorState
        self.monitorState = False
        return out

    def getControl(self):
        """
        returns a dictionary of control information for a PV
        Example::
        
        >>> pv = epicsPV('13IDC:m1')
        >>> for field, value in pv.getControl().items():
        >>>    print field, ':', value
        status  :  0
        severity  :  0
        precision  :  5
        units  :  mm
        lower_alarm_limit  :  0.0
        upper_alarm_limit  :  0.0
        lower_warning_limit  :  0.0
        upper_warning_limit  :  0.0
        lower_disp_limit  :  -2.4
        upper_disp_limit  :  2.4
        upper_ctrl_limit  :  2.4
        lower_ctrl_limit  :  -2.4
        """
        epics.poll()
        return self.get_ctrlvars()
      
    def array_get(self, count=None):
        """ returns PV value    """
        return self.getw(count=count)

    def getw(self, count=None):
        """ returns PV value"""
        return self.get(count=count)

    def getValue(self):
        """  get most recent value for PV   """
        return self.get()
    
    def putw(self, value, wait=False):
        """ set PV value"""
        self.put(value, wait=wait)
        
    def putWait(self, value):
        """ put PV value, waits for the callback to
        occur before it returns.  """
        self.put(value, wait=True)
      



from .. import Device

class AD_ImagePlugin(Device):
    """
    AreaDetector Image Plugin
    """
    attrs = ('ArrayData',
             'UniqueId', 'UniqueId_RBV',
             'NDimensions', 'NDimensions_RBV',
             'ArraySize0', 'ArraySize0_RBV',
             'ArraySize1', 'ArraySize1_RBV',
             'ArraySize2', 'ArraySize2_RBV',
             'ColorMode', 'ColorMode_RBV')

    _nonpvs = ('_prefix', '_pvs', '_delim')

    def __init__(self, prefix):
        Device.__init__(self, prefix, delim='', mutable=False,
                        attrs=self.attrs)

    def ensure_value(self, attr, value, wait=False):
        """ensures that an attribute with an associated _RBV value is
        set to the specifed value
        """
        rbv_attr = "%s_RBV" % attr
        if rbv_attr not in self._pvs:
            return self._pvs[attr].put(value, wait=wait)

        if  self._pvs[rbv_attr].get(as_string=True) != value:
            self._pvs[attr].put(value, wait=wait)




#
# example of using a connection callback that will be called
# for any change in connection status

import epics
import time
import sys
from  pvnames import motor1

write = sys.stdout.write
def onConnectionChange(pvname=None, conn= None, **kws):
    write('PV connection status changed: %s %s\n' % (pvname,  repr(conn)))
    sys.stdout.flush()

def onValueChange(pvname=None, value=None, host=None, **kws):
    write('PV value changed: %s (%s)  %s\n' % ( pvname, host, repr(value)))
    sys.stdout.flush()
mypv = epics.PV(motor1, 
                connection_callback= onConnectionChange,
                callback= onValueChange)

mypv.get()

write('Now waiting, watching values and connection changes:\n')
t0 = time.time()
while time.time()-t0 < 300:
    time.sleep(0.01)



#!/usr/bin/env python
#
# test simulator for testing pyepics.
#
# this script changes values definied in the pydebug.db, which all
# the tests scripts use.  This script must be running (somewhere)
# for many of the tests (callbacks, etc) to work.

#
import epics
import time
import random
import numpy

prefix = 'PyTest:'

global NEEDS_INIT

NEEDS_INIT = True
SLEEP_TIME = 0.10

def onConnect(pvname=None, conn=None, **kws):
    global NEEDS_INIT
    NEEDS_INIT = conn


def make_pvs(*args, **kwds):
    # print "Make PVS '  ", prefix,  args
    # print  [("%s%s" % (prefix, name)) for name in args]
    pvlist = [epics.get_pv("%s%s" % (prefix, name)) for name in args]
    for pv in pvlist:
        pv.connection_callbacks.append(onConnect)
    return pvlist


mbbos    = make_pvs("mbbo1","mbbo2")
pause_pv = make_pvs("pause",)[0]
longs    = make_pvs("long1", "long2", "long3", "long4")
strs     = make_pvs("str1", "str2")
analogs  = make_pvs("ao1", "ai1", "ao2", "ao3")
binaries = make_pvs("bo1", "bi1")

char_waves   = make_pvs("char128", "char256", "char2k", "char64k")
double_waves = make_pvs("double128", "double2k", "double64k")
long_waves   = make_pvs("long128", "long2k", "long64k")
str_waves    = make_pvs("string128", "string2k", "string64k")
subarrays    = make_pvs("subArr1", "subArr2", "subArr3", "subArr4" )
subarray_driver = make_pvs("wave_test",)[0]


def initialize_data():
    subarray_driver.put(numpy.arange(64)/12.0)
    for p in mbbos:
        p.put(1)

    for i, p in enumerate(longs):
        p.put((i+1))

    for i, p in enumerate(strs):
        p.put("String %s" % (i+1))

    for i, p in enumerate(binaries):
        p.put((i+1))

    for i, p in enumerate(analogs):
        p.put((i+1)*1.7135000 )

    epics.caput('%sao1.EGU' % prefix, 'microns')
    epics.caput('%sao1.PREC' % prefix, 4)
    epics.caput('%sai1.PREC' % prefix, 2)
    epics.caput('%sao2.PREC' % prefix, 3)

    char_waves[0].put([60+random.randrange(30) for i in range(128)])
    char_waves[1].put([random.randrange(256) for i in range(256)])
    char_waves[2].put([random.randrange(256) for i in range(2048)])
    char_waves[3].put([random.randrange(256) for i in range(65536)])


    long_waves[0].put([i+random.randrange(2) for i in range(128)])
    long_waves[1].put([i+random.randrange(128) for i in range(2048)])
    long_waves[2].put([i  for i in range(65536)])

    double_waves[0].put([i+random.randrange(2) for i in range(128)])
    double_waves[1].put([random.random() for i in range(2048)])
    double_waves[2].put([random.random() for i in range(65536)])

    pause_pv.put(0)
    str_waves[0].put([" String %i" % (i+1) for i in range(128)])
    print( 'Data initialized')

text = '''line 1
this is line 2
and line 3
here is another line
this is the 5th line
line 6
line 7
line 8
line 9
line 10
line 11
'''.split('\n')

start_time = time.time()
count = 0
long_update = 0
lcount =1

while True:
    if NEEDS_INIT:
        initialize_data()
        time.sleep(SLEEP_TIME)
        NEEDS_INIT = False

    time.sleep(SLEEP_TIME)

    count = count + 1
    if count  == 3: print( 'running')
    if count > 99999999: count = 1

    t0 = time.time()
    if pause_pv.get() == 1:
        # pause for up to 300 seconds if pause was selected
        t0 = time.time()
        while time.time()-t0 < 300:
            time.sleep(SLEEP_TIME)
            if pause_pv.get() == 0:
                break
        pause_pv.put(0)
    noise = numpy.random.normal

    analogs[0].put( 100*(random.random()-0.5))
    analogs[1].put( 76.54321*(time.time()-start_time))
    analogs[2].put( 0.3*numpy.sin(time.time() / 2.302) + noise(scale=0.4)  )
    char_waves[0].put([45+random.randrange(64) for i in range(128)])

    if count % 3 == 0:
        analogs[3].put( numpy.exp((max(0.001,  noise(scale=0.03)
                                       + numpy.sqrt((count/16.0) % 87.)))))

        long_waves[1].put([i+random.randrange(128) for i in range(2048)])
        str_waves[0].put(["Str%i_%.3f" % (i+1, 100*random.random()) for i in range(128)])

    if t0-long_update >= 1.0:
        long_update=t0
        lcount = (lcount + 1) % 10
        longs[0].put(lcount)
        char_waves[1].put(text[lcount])
        double_waves[2].put([random.random() for i in range(65536)])
        double_waves[1].put([random.random() for i in range(2048)])



"""This script tests using EPICS CA and Python threads together

Based on code from  Friedrich Schotte, NIH, modified by Matt Newville
20-Apr-2010
"""

import time
from threading import Thread
import epics
import sys
import pvnames

epics.caput(pvnames.motor1,   0.3)
epics.caput(pvnames.motor2,  -1.0)
time.sleep(1.0)
epics.caput(pvnames.motor2, 1.0, wait=True)
sys.stdout.write('done with initial moves.\n')

def run_test(pvname,  target, run_name='thread c'):
    sys.stdout.write( ' -> thread  "%s"\n' % run_name)
    def onChanges(pvname=None, value=None, char_value=None, **kw):
        sys.stdout.write('      %s = %s (%s)\n' % (pvname, char_value, run_name))
    epics.ca.context_create()
    p = epics.PV(pvname)
    sys.stdout.write('Put %s to %.3f   (%s)\n' % (pvname, target,run_name))
    p.put(target, wait=True)
    sys.stdout.write( 'Done with Thread %s\n' % run_name)
    epics.ca.context_destroy()
    
epics.ca.show_cache()

sys.stdout.write( "Run 2 Background Threads doing simultaneous put-with-waits:\n")
th1 = Thread(target=run_test,args=(  pvnames.motor1,  0.5,  'A'))
th2 = Thread(target=run_test,args=(  pvnames.motor2,  .22,  'B'))
th1.start()
th2.start()

epics.ca.show_cache()
th2.join()
th1.join()

sys.stdout.write( 'Done.\n')
time.sleep(0.01)
epics.ca.show_cache()



#!/usr/bin/python

# test of simplest device

from epics.devices import ao
import sys
import time
from epics.utils import IOENCODING

def test_aodevice():
    myao = ao('PyTest:ao1')
    assert len(myao._pvs) > 10
    assert myao.VAL is not None
    myao.write_state('tmp_aostate.txt')

    time.sleep(0.5)
    flines = []
    with open('tmp_aostate.txt', encoding=IOENCODING) as fh:
        flines.extend( fh.readlines())
    assert len(flines) > 10



import time
import pytest
import subprocess
from tempfile import NamedTemporaryFile as NTF
import epics


cas_test_db = '''
                record(ao, "test:ao") {
                    field(ASG, "rps_threshold")
                    field(DRVH, "10")
                    field(DRVL, "0")
                }

                record(bo, "test:bo") {
                    field(ASG, "rps_lock")
                    field(ZNAM, "OUT")
                    field(ONAM, "IN")
                }

                record(ao, "test:ao2") {
                    field(DRVH, "5")
                    field(DRVL, "1")
                }

                record(bo, "test:permit") {
                    field(VAL, "0")
                    field(PINI, "1")
                    field(ZNAM, "DISABLED")
                    field(ONAM, "ENABLED")
                }
            '''

cas_rules = '''
                ASG(DEFAULT) {
                    RULE(1,READ)
                    RULE(1,WRITE,TRAPWRITE)
                }

                ASG(rps_threshold) {
                    INPA("$(P):permit")
                    RULE(1, READ)
                    RULE(0, WRITE, TRAPWRITE) {
                        CALC("A=1")
                    }
                    RULE(1, WRITE, TRAPWRITE) {
                        CALC("A=0")
                    }
                }

                ASG(rps_lock) {
                    INPA("$(P):permit")
                    RULE(1, READ)
                    RULE(1, WRITE, TRAPWRITE) {
                        CALC("A=0")
                    }
                }
            '''

# use yield_fixture() for compatibility with pytest < 2.10
@pytest.yield_fixture(scope='module')
def softioc():
    with NTF(mode='w+') as cf, NTF(mode='w+') as df:
        cf.write(cas_rules)
        cf.flush()
        df.write(cas_test_db)
        df.flush()

        proc = subprocess.Popen(['softIoc', '-D',
                                 '/home/travis/mc/envs/testenv/epics/dbd/softIoc.dbd',
                                 '-m', 'P=test', '-a', cf.name,
                                 '-d', df.name],
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE)
        yield proc

        try:
            proc.kill()
            proc.wait()
        except OSError:
            pass


@pytest.yield_fixture(scope='module')
def pvs():
    pvlist = ['test:ao', 'test:ao.DRVH', 'test:bo', 'test:ao2',
              'test:permit']
    pvs = dict()
    for name in pvlist:
        pv = epics.get_pv(name)
        pv.wait_for_connection()
        pvs[pv.pvname] = pv

    yield pvs
    for pv in pvs.values():
        pv.disconnect()


def test_connected(softioc, pvs):
    for pv in pvs.values():
        assert pv.connected

def test_permit_disabled(softioc, pvs):
    # with the permit disabled, all test pvs should be readable/writable
    for pv in pvs.values():
        assert pv.read_access and pv.write_access

def test_permit_enabled(softioc, pvs):
    # set the run-permit
    pvs['test:permit'].put(1, wait=True)
    assert pvs['test:permit'].get(as_string=True, use_monitor=False) == 'ENABLED'

    # rps_lock rule should disable write access
    assert pvs['test:bo'].write_access is False
    with pytest.raises(epics.ca.CASeverityException):
        pvs['test:bo'].put(1, wait=True)

    # rps_threshold rule should disable write access to metadata, not VAL
    assert pvs['test:ao'].write_access is True
    assert pvs['test:ao.DRVH'].write_access is False
    with pytest.raises(epics.ca.CASeverityException):
        pvs['test:ao.DRVH'].put(100, wait=True)

def test_pv_access_event_callback(softioc, pvs):
    # clear the run-permit
    pvs['test:permit'].put(0, wait=True)
    assert pvs['test:permit'].get(as_string=True, use_monitor=False) == 'DISABLED'

    def lcb(read_access, write_access, pv=None):
        assert pv.read_access == read_access
        assert pv.write_access == write_access
        pv.flag = True

    bo = epics.get_pv('test:bo', access_callback=lcb)
    bo.flag = False

    # set the run-permit to trigger an access rights event
    pvs['test:permit'].put(1, wait=True)
    assert pvs['test:permit'].get(as_string=True, use_monitor=False) == 'ENABLED'

    assert bo.flag is True
    bo.access_callbacks = []


def test_ca_access_event_callback(softioc, pvs):
    # clear the run-permit
    pvs['test:permit'].put(0, wait=True)
    assert pvs['test:permit'].get(as_string=True, use_monitor=False) == 'DISABLED'

    bo_id = epics.ca.create_channel('test:bo')
    assert bo_id is not None

    def lcb(read_access, write_access):
        assert read_access and write_access
        lcb.sentinal = True

    lcb.sentinal = False
    epics.ca.replace_access_rights_event(bo_id, callback=lcb)

    assert lcb.sentinal is True
    epics.ca.clear_channel(bo_id)


def test_connection_callback(softioc, pvs):
    results = []

    def callback(conn, **kwargs):
        results.append(conn)

    pv = epics.PV('test:ao', connection_callback=callback)
    pv.wait_for_connection()
    softioc.kill()
    softioc.wait()

    t0 = time.time()
    while pv.connected and (time.time() - t0) < 5:
        time.sleep(0.1)

    assert True in results
    assert False in results



import time
import epics
import sys
import pvnames

# Test than when a PV connections all callbacks fire successfully
#
# does not require Setup/simulator.py to be running (PV is deliberately one which does not change)

write = sys.stdout.write


def wait(timeout=10):
    t0 = time.time()
    while time.time() - t0 < timeout:
        time.sleep(1.e-4)

got_callback_a = False
got_callback_b = False

def callback_a(pvname=None, value=None, **kw):
    global got_callback_a
    write( "Got callback A (%s, %s)\n" % (pvname, repr(value)) )
    got_callback_a = True

def callback_b(pvname=None, value=None, **kw):
    global got_callback_b
    write( "Got callback B (%s, %s)\n" % (pvname, repr(value)) )
    got_callback_b = True

def CB1(pvname=None, value=None, char_value=None, **kw):
    print( 'CB1 PV Changed! ', pvname, value, char_value )

def CB2(pvname=None, value=None, char_value=None, **kw):
    print( 'CB2 ! ', pvname, value, char_value)


def test_initial_callbacks():
    global got_callback_a, got_callback_b
    pvname = pvnames.non_updating_pv
    mypv = epics.PV(pvname, callback=(callback_a, callback_b))

    wait(3)
    assert mypv.connected
    assert got_callback_a
    assert got_callback_b

def test_multiple_callbacks():
    global got_callback_a, got_callback_b
    got_callback_a =  got_callback_b = False
    mypv = epics.PV(pvnames.double_pv)
    mypv.add_callback(callback_a)

    wait(2)
    assert got_callback_a
    assert not got_callback_b

    mypv.add_callback(callback_b)
    wait(2)
    assert got_callback_b


if __name__ == '__main__':
    test_initial_callbacks()
    test_multiple_callbacks()



#!/usr/bin/env python
from .. import Device

class bi(Device):
    """
    Simple binary input device
    """

    attrs = ('INP', 'ZNAM', 'ONAM', 'RVAL', 'VAL', 'EGU', 'HOPR', 'LOPR',
               'PREC', 'NAME', 'DESC', 'DTYP')

    def __init__(self, prefix, **kwargs):
        if prefix.endswith('.'):
            prefix = prefix[:-1]
        Device.__init__(self, prefix, delim='.', attrs=self.attrs, **kwargs)





"""This script tests using EPICS CA and Python threads together
Based on code from  Friedrich Schotte, NIH, modified by Matt Newville
19-Apr-2010
"""
import time
from  sys import stdout
from threading import Thread
import epics
from epics.ca import CAThread

from  pvnames import updating_pvlist
pvlist_a = updating_pvlist[:-1]
pvlist_b = updating_pvlist[1:]

def run_test(runtime=1, pvnames=None,  run_name='thread c'):
    msg = '-> thread "%s" will run for %.3f sec, monitoring %s\n'
    stdout.write(msg % (run_name, runtime, pvnames))
    def onChanges(pvname=None, value=None, char_value=None, **kw):
        stdout.write('   %s = %s (%s)\n' % (pvname, char_value, run_name))
        stdout.flush()

    # epics.ca.use_initial_context()   #  epics.ca.create_context()
    start_time = time.time()
    pvs = [epics.PV(pvn, callback=onChanges) for pvn in pvnames]

    while time.time()-start_time < runtime:
        time.sleep(0.1)

    [p.clear_callbacks() for p in pvs]
    stdout.write( 'Completed Thread  %s\n' % ( run_name))

stdout.write( "First, create a PV in the main thread:\n")
p = epics.PV(updating_pvlist[0])

stdout.write("Run 2 Background Threads simultaneously:\n")
th1 = CAThread(target=run_test,args=(3, pvlist_a,  'A'))
th1.start()

th2 = CAThread(target=run_test,args=(6, pvlist_b, 'B'))
th2.start()

th2.join()
th1.join()
stdout.write('Done\n')



#!/usr/bin/python

# test of simplest device
from epics import PV, get_pv, caget, camonitor, camonitor_clear

import os
import psutil
import time

import pvnames
mypv = pvnames.updating_pv1


def test_connect_disconnect():
    pv = PV(mypv, auto_monitor=True, callback=lambda **args: ...)

    pv.wait_for_connection()

    # check that PV is connected
    assert pv.connected is True

    # check that data is received
    value = pv.get()
    assert value is not None

    pv.disconnect()

    # check that PV is disconnected
    assert pv.connected is False

    # check that no data is received after disconnect
    value = pv.get()
    assert value is None


def test_reconnect():
    # connect and disconnect
    pv = PV(mypv, auto_monitor=True, callback=lambda **args: ...)
    pv.wait_for_connection()
    pv.disconnect()

    # try to reconnect to the same PV
    connected = pv.reconnect()

    # check that PV is connected
    assert connected is True

    # check that data is received
    value = pv.get()
    assert value is not None


def test_with_two_PVs():
    # create 2 PV objects connecting to the same PV signal
    pv1 = PV(mypv, auto_monitor=True, callback=lambda **args: ...)
    pv2 = PV(mypv, auto_monitor=True, callback=lambda **args: ...)

    pv1.wait_for_connection()
    pv2.wait_for_connection()

    # check that both PVs are connected
    assert pv1.connected is True
    assert pv2.connected is True

    # check that data is received
    assert pv1.get() is not None
    assert pv2.get() is not None

    # disconnect 1 PV
    pv1.disconnect()

    # check that the first PV is disconnected and doesn't receive data
    assert pv1.connected is False
    assert pv1.get() is None

    # check that the other PV is connected and still receives data
    assert pv2.connected is True
    assert pv2.get() is not None


def test_with_PV_and_getPV():
    # create 2 PV objects connecting to the same PV signal, one using PV class and the other one using get_pv()
    pv1 = PV(mypv, auto_monitor=True, callback=lambda **args: ...)
    pv2 = get_pv(mypv)

    pv1.wait_for_connection()
    pv2.wait_for_connection()

    # check that both PVs are connected
    assert pv1.connected is True
    assert pv2.connected is True

    # check that data is received
    assert pv1.get() is not None
    assert pv2.get() is not None

    # disconnect 1 PV
    pv1.disconnect()

    time.sleep(1)

    # check that the first PV is disconnected and doesn't receive data
    assert pv1.connected is False
    assert pv1.get() is None

    # check that the other PV is connected and still receives data
    assert pv2.connected is True
    assert pv2.get() is not None


def test_with_getPV():
    # create 2 PV objects connecting to the same PV signal using get_pv()
    pv1 = get_pv(mypv)
    pv2 = get_pv(mypv)

    pv1.wait_for_connection()
    pv2.wait_for_connection()

    # check that both PVs are connected
    assert pv1.connected is True
    assert pv2.connected is True

    # check that data is received
    assert pv1.get() is not None
    assert pv2.get() is not None

    # disconnect 1 PV
    pv1.disconnect()

    time.sleep(1)

    # check that the first PV is disconnected and doesn't receive data
    assert pv1.connected is False
    assert pv1.get() is None

    # check that the other PV is also disconnected and doesn't receive data either
    assert pv2.connected is False
    assert pv2.get() is None


def test_with_caget():
    pv = PV(mypv, auto_monitor=True, callback=lambda **args: ...)
    pv.wait_for_connection()

    # check that the PV is connected and  data is received
    assert pv.connected is True
    assert pv.get() is not None

    # use caget to get data from the same PV
    assert caget(mypv) is not None

    # disconnect PV object
    pv.disconnect()

    # check that the PV is disconnected and doesn't receive data
    assert pv.connected is False
    assert pv.get() is None

    # check that you can still use caget to get data from the same PV
    assert caget(mypv) is not None


def test_with_caget_nomonitor():
    pv = PV(mypv, auto_monitor=True, callback=lambda **args: ...)
    pv.wait_for_connection()

    # check that the PV is connected and  data is received
    assert pv.connected is True
    assert pv.get() is not None

    # use caget to get data from the same PV
    assert caget(mypv, use_monitor=False) is not None

    # disconnect PV object
    pv.disconnect()

    # check that the PV is disconnected and doesn't receive data
    assert pv.connected is False
    assert pv.get() is None

    # check that you can still use caget to get data from the same PV
    assert caget(mypv, use_monitor=False) is not None


def test_with_camonitor():
    pv = PV(mypv, auto_monitor=True, callback=lambda **args: ...)
    pv.wait_for_connection()

    # check that the PV is connected and  data is received
    assert pv.connected is True
    assert pv.get() is not None

    # use camonitor
    received = {'flag': False}

    def callback(**args):
        received['flag'] = True
    camonitor(mypv, callback=callback)
    time.sleep(1)

    # check that the monitor receives data
    assert received['flag'] is True

    # disconnect PV object
    pv.disconnect()
    time.sleep(1)

    # check that the PV is disconnected and doesn't receive data
    assert pv.connected is False
    assert pv.get() is None

    # reset the flag to check that new data is received by camonitor
    received['flag'] = False
    time.sleep(1)
    assert received['flag'] is True

    # clear the monitor
    camonitor_clear(mypv)
    time.sleep(1)

    # reset the flag to check that no new data is received by camonitor
    received['flag'] = False
    time.sleep(1)
    assert received['flag'] is False


def test_memleak_disconnect():
    # try to connect multiple times to the same PV
    mem = []
    for i in range(int(2)):
        for j in range(int(1000)):
            pv = PV(mypv, auto_monitor=True, callback=lambda **args: ...)
            pv.disconnect()

        process = psutil.Process(os.getpid())
        mem.append(process.memory_info().rss)

    # check used memory by the process didn't increase by more than 1%
    assert mem[1]/mem[0] < 1.01



from __future__ import print_function

import sys
import time
import epics
import pvnames

pvlist = (
    pvnames.str_pv,
    pvnames.int_pv,
    pvnames.float_pv,
    pvnames.enum_pv,
    pvnames.char_arr_pv,
    pvnames.long_pv,
    pvnames.long_arr_pv,
    pvnames.double_pv,
    pvnames.double_arr_pv,
    pvnames.string_arr_pv,
    )


def RunTest(pvlist, use_preempt=True, maxlen=16384,
            use_numpy=True, use_time=False, use_ctrl=False):
    msg= ">>>Run Test: %i pvs, numpy=%s, time=%s, ctrl=%s, preempt=%s"
    print( msg % (len(pvlist), use_numpy, use_time, use_ctrl, use_preempt))

    epics.ca.HAS_NUMPY =  epics.ca.HAS_NUMPY and use_numpy
    epics.ca.PREEMPTIVE_CALLBACK = use_preempt
    epics.ca.AUTOMONITOR_MAXLENGTH = maxlen
    chids= []
    epics.ca.initialize_libca()

    def onConnect(pvname=None,  **kw):
        print(' on Connect %s %s' % (pvname, repr(kw)))

    def onChanges(chid=None, value=None, **kw):
        print(' on Change chid=%i value=%s' % (int(chid), repr(value)))

    for pvname in pvlist:
        chid = epics.ca.create_channel(pvname, callback=onConnect)
        epics.ca.connect_channel(chid)
        eventID = epics.ca.create_subscription(chid, callback=onChanges)
        chids.append((chid, eventID))
        epics.poll(evt=0.025, iot=5.0)
    epics.poll(evt=0.025, iot=10.0)
    time.sleep(0.05)
    for (chid, eventID) in chids:
        print('=== %s   chid=%s' % (epics.ca.name(chid), repr(chid)))
        time.sleep(0.005)
        ntype = epics.ca.promote_type(chid, use_ctrl=use_ctrl,
                                      use_time=use_time)
        val  = epics.ca.get(chid, ftype=ntype)
        cval = epics.ca.get(chid, as_string=True)
        if epics.ca.element_count(chid) > 10:
            val = val[:10]
        time.sleep(0.005)
        print("%i %s %s" % (ntype, epics.dbr.Name(ntype).lower(), cval))
    time.sleep(0.5)
    print('----- finalizing CA')
    epics.ca.finalize_libca()
    time.sleep(0.05)


def test1():
    RunTest(pvlist, use_preempt=True, use_numpy=True, use_time=False, use_ctrl=False)
    RunTest(pvlist, use_preempt=True, use_numpy=True, use_time=False, use_ctrl=True)
    RunTest(pvlist, use_preempt=True, use_numpy=True, use_time=True, use_ctrl=False)

def test2():
    RunTest(pvlist, use_preempt=True, use_numpy=False, use_time=False, use_ctrl=False)
    RunTest(pvlist, use_preempt=True, use_numpy=False, use_time=False, use_ctrl=True)
    RunTest(pvlist, use_preempt=True, use_numpy=False, use_time=True, use_ctrl=False)

def test3():
    RunTest(pvlist, use_preempt=False, use_numpy=True, use_time=False, use_ctrl=False)
    RunTest(pvlist, use_preempt=False, use_numpy=True, use_time=False, use_ctrl=True)
    RunTest(pvlist, use_preempt=False, use_numpy=True, use_time=True, use_ctrl=False)

def test4():
    RunTest(pvlist, use_preempt=False, use_numpy=False, use_time=False, use_ctrl=False)
    RunTest(pvlist, use_preempt=False, use_numpy=False, use_time=False, use_ctrl=True)
    RunTest(pvlist, use_preempt=False, use_numpy=False, use_time=True, use_ctrl=False)



import sys
import time
import epics
import pvnames

pvn = pvnames.alarm_pv
pvn = pvnames.alarm_comp
pvn = pvnames.alarm_trippoint

def alarmHandler(pvname=None,value=None,char_value=None,
                 comparison=None,trip_point=None,**kw):
    sys.stdout.write( 'Alarm! %s at %s ! \n' %( pvname,  time.ctime()))
    sys.stdout.write( 'Alarm  Comparison =%s  \n' %( comparison))
    sys.stdout.write( 'Alarm  TripPoint      =%s  \n' %( repr(trip_point)))
    sys.stdout.write( 'Current Value         =%s  \n' %(char_value))    

epics.Alarm(pvname = pvnames.alarm_pv,
            comparison = pvnames.alarm_comp,
            trip_point    =pvnames.alarm_trippoint,
            callback = alarmHandler,
            alert_delay=5.0)


t0 = time.time()

sys.stdout.write('Waiting for pv %s to change! \n' % pvnames.alarm_pv)
sys.stdout.write('Alarm settings:  comp=%s,  trip_point=%s\n' % (pvnames.alarm_comp,
                                                                 pvnames.alarm_trippoint))
sys.stdout.write('You may have to make this happen!!\n')

while time.time()-t0 < 30:
    try:
        epics.ca.poll()
    except KeyboardInterrupt:
        break



# tests caget (and so pv.get() and ca.get()) for "large arrays"
# especially for "slow networks".  
# the 3 pvs listed here are:   
#   13GEXMAP:mca1:                2048 ints (simple MCA record)
#   GSE-PIL1:image1:ArrayData    94965 longs (Pilatus 100k) 
#   13MARCCD1:image1:ArrayData  420000 ints  (MAR165)
# obviously, you'll need to alter these names
#
# TIMEOUT sets the pend_io() time.  If working with large
# arrays and/or slow networks, you may want to test for a
# suitable safe timeout.
import epics
import time
pvnames = ('13GEXMAP:mca1',
           'GSE-PIL1:image1:ArrayData',
           '13MARCCD2:image1:ArrayData',
           )

TIMEOUT = 15.0  # Timeout in seconds

for name in pvnames:
    t0 = time.time()
    value = epics.caget(name, timeout=TIMEOUT)
    dt = time.time()-t0
    if value is None:
        print( 'cannot get value for ', name)
    else:
        print( "%s: npts=%i, sum=%i, max=%i, time=%.3fs" % (name,
                                                           len(value),
                                                           value.sum(),
                                                           value.max(), dt))

print( 'done.')


